#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* input file */
#define X1NAME "hattack.bin"
#define PCGNAME "pcg.txt"
#define PATCHNAME "patch.bin"
#define LOADERNAME "loader.bin"

/* output file */
#define TAPENAME "hover2.p6"
#define DSKNAME "hover.dsk"
#define D88NAME "hover.d88"

/* data address and size */
#define OFFSET 0xc000
#define SIZE_IN 0x1430
#define SIZE_OUT 0x3000

/* for d88 */
#define HEADER_SIZE 0x2b0
#define NHEAD 1
#define NTRACK 40
#define NSECTOR 16
#define SECTOR_SIZE 0x100
#define TRACK_DENSITY 1

void header(FILE *fp);
void secthead(int c, int h, int r, int n, FILE *fp);
void copy(const char *fname, int *c, int *h, int *r, int *n, FILE *fp_out);
void fput2(int data, FILE *fp);
void fput4(long data, FILE *fp);

int main()
{
  FILE *fp;
  int i, j;
  int size, address;
  unsigned char x1data[OFFSET + SIZE_IN], p6data[OFFSET + SIZE_OUT];
  unsigned int pcgdata;
  char line[1000];
  int c, h, r, n;

  /* read x1 file */
  fp = fopen(X1NAME, "rb");
  if (fp == NULL) {
    printf("cannot open %s for reading\n", X1NAME);
    exit(1);
  }
  if (fread(&x1data[OFFSET], 1, SIZE_IN, fp) < SIZE_IN) {
    printf("%s size is too small\n", X1NAME);
    exit(1);
  }
  fclose(fp);
  memset(p6data, 0, sizeof(p6data));
  memcpy(p6data, x1data, sizeof(x1data));

  /* convert graphic data */
  fp = fopen(PCGNAME, "r");
  if (fp == NULL) {
    printf("cannot open %s for reading\n", PCGNAME);
    exit(1);
  }
  while (fgets(line, sizeof(line), fp)) {
    if (strstr(line, "&H") == NULL) {
      continue;
    }
    sscanf(strstr(line, "&H") + 2, "%x", &i);
    for (j = 0; j < 8; j++) {
      /* blue */
      sscanf(strchr(line, '"') + 1 + j * 2, "%02x", &pcgdata);
      p6data[0xe000 + j * 2 + (i & 0xf0) + ((i & 0x0f) << 8)] = pcgdata;
      /* red */
      sscanf(strchr(line, '"') + 1 + (8 + j) * 2, "%02x", &pcgdata);
      p6data[0xe001 + j * 2 + (i & 0xf0) + ((i & 0x0f) << 8)] = pcgdata;
      /* green */
      sscanf(strchr(line, '"') + 1 + (16 + j) * 2, "%02x", &pcgdata);
      p6data[0xe000 + j * 2 + (i & 0xf0) + ((i & 0x0f) << 8)] |= pcgdata;
    }
  }
  fclose(fp);

  /* patch program */
  fp = fopen(PATCHNAME, "rb");
  if (fp == NULL) {
    printf("cannot open patch.bin for reading\n");
    exit(1);
  }
  while ((size = fgetc(fp)) != EOF) {
    size += fgetc(fp) << 8;
    address = fgetc(fp) + (fgetc(fp) << 8);
    for (i = 0; i < size; i++) {
      p6data[OFFSET + address + i] = fgetc(fp);
    }
  }
  fclose(fp);

  /* write tape image */
  fp = fopen(TAPENAME, "wb");
  if (fp == NULL) {
    printf("cannot open %s for writing\n", TAPENAME);
    exit(1);
  }
  for (i = 0; i < 10; i++) {
    fputc(0xd3, fp);
  }
  fputs("MAIN", fp);
  fputc(0, fp);
  fputc(0, fp);
  fwrite(&p6data[OFFSET], 1, SIZE_OUT, fp);
  fclose(fp);
  printf("output to %s\n", TAPENAME);

  /* write disk image */
  fp = fopen(LOADERNAME, "rb");
  if (fp == NULL) {
    printf("cannot open %s for reading\n", LOADERNAME);
    exit(1);
  }
  memset(p6data, 0xff, 0x1000);
  fread(p6data, 1, 0x100, fp);
  fclose(fp);

  fp = fopen(DSKNAME, "wb");
  if (fp == NULL) {
    printf("cannot open %s for writing\n", DSKNAME);
    exit(1);
  }
  fwrite(p6data, 1, 0x1000, fp);
  fwrite(&p6data[OFFSET], 1, SIZE_OUT, fp);
  for (i = 0x1000 + SIZE_OUT; i < 0x1000 * 40;i++) {
    fputc(0xff, fp);
  }
  fclose(fp);
  printf("output to %s\n", DSKNAME);

  fp = fopen(D88NAME, "wb");
  if (fp == NULL) {
    printf("cannot open %s for writing\n", D88NAME);
    exit(1);
  }
  header(fp);
  c = 0;
  h = 0;
  r = 1;
  n = 1;
  copy(DSKNAME, &c, &h, &r, &n, fp);
  for (; c < NTRACK; c++) {
    for (; h < NHEAD; h++) {
      for (; r <= NSECTOR; r++) {
	secthead(c, h, r, n, fp);
	for (i = 0; i < SECTOR_SIZE; i++) {
	  fputc(0xff, fp);  /* no data */
	}
      }
      r = 1;
    }
    h = 0;
  }
  fclose(fp);
  printf("output to %s\n", D88NAME);

  return 0;
}

void header(FILE *fp)
{
  int i, j;
  long size, address;

  /* name(17), reserved(9), protect(1) */
  for (i = 0; i < 27; i++) {
    fputc(0x00, fp);
  }

  /* track density */
  switch (TRACK_DENSITY) {
  case 1:
    fputc(0x00, fp);
    break;
  case 2:
    fputc(0x10, fp);
    break;
  default:
    fputc(0x20, fp);
    break;
  }

  /* disk size */
  size = HEADER_SIZE + NHEAD * NTRACK * NSECTOR * (SECTOR_SIZE + 16);
  fput4(size, fp);

  /* track data table */
  address = HEADER_SIZE;
  for (i = 0; i < 82; i++) {  /* track */
    for (j = 0; j < 2; j++) {  /* head */
      if (i < NTRACK && j < NHEAD) {
	fput4(address, fp);
	address += NSECTOR * (SECTOR_SIZE + 16);
      }else{
	fput4(0, fp);
      }
    }
  }
}

void secthead(int c, int h, int r, int n, FILE *fp)
{
  fputc(c, fp);		/* track */
  fputc(h, fp);		/* head */
  fputc(r, fp);		/* sector */
  fputc(n, fp);		/* sector size=128*2^n */
  fput2(NSECTOR, fp);	/* sectors in track */
  fputc(0, fp);		/* double density */
  fputc(0, fp);		/* not deleted */
  fputc(0, fp);		/* status normal */
  fputc(0, fp);		/* reserved */
  fputc(0, fp);		/* reserved */
  fputc(0, fp);		/* reserved */
  fputc(0, fp);		/* reserved */
  fputc(0, fp);		/* reserved */
  fput2(SECTOR_SIZE, fp);
}


void copy(const char *fname, int *c, int *h, int *r, int *n, FILE *fp_out)
{
  FILE *fp_in;
  unsigned char data[SECTOR_SIZE];

  fp_in = fopen(fname, "rb");
  if (fp_in == NULL) {
    printf("cannot open %s for reading\n", fname);
    exit(1);
  }

  while((fread(data, 1, SECTOR_SIZE, fp_in)) == SECTOR_SIZE) {
  if (*c > NTRACK - 1 || *h > NHEAD - 1 || *r > NSECTOR) {
    printf("disk full\n");
    exit(1);
  }

  secthead(*c, *h, *r, *n, fp_out);
  fwrite(data, 1, SECTOR_SIZE, fp_out);
  (*r)++;

  if (*r > NSECTOR) {
    (*r) = 1;
    (*h)++;
    if (*h > NHEAD - 1) {
      (*h) = 0;
      (*c)++;
    }
  }
 }

  fclose(fp_in);
}

void fput2(int data, FILE *fp)
{
  fputc(data & 0xff, fp);
  fputc((data >> 8) & 0xff, fp);
}

void fput4(long data, FILE *fp)
{
  fputc(data & 0xff, fp);
  fputc((data >> 8) & 0xff, fp);
  fputc((data >> 16) & 0xff, fp);
  fputc((data >> 24) & 0xff, fp);
}
